/*
** 2013-07-07
** MipMappedText
**
** This demonstration of concept builds on the Molehill tutorials of
** Moussa Dembélé -- http://mousman.com/agal/?p=621
** and
** Marco Scabia -- http://www.adobe.com/devnet/flashplayer/articles/mipmapping.html
** Improved camera controls from
** Jackson Dunstan -- http://jacksondunstan.com/articles/1655
** are then augmented and merged in.
*/
package
{
import com.adobe.utils.AGALMiniAssembler;

import flash.display.Bitmap;
import flash.display.BitmapData;
import flash.display.BlendMode;
import flash.display.Shader;
import flash.display.Sprite;
import flash.display.StageAlign;
import flash.display.StageQuality;
import flash.display.StageScaleMode;
import flash.display3D.Context3D;
import flash.display3D.Context3DProgramType;
import flash.display3D.Context3DTextureFormat;
import flash.display3D.Context3DVertexBufferFormat;
import flash.display3D.IndexBuffer3D;
import flash.display3D.Program3D;
import flash.display3D.VertexBuffer3D;
import flash.display3D.textures.Texture;
import flash.events.Event;
import flash.filters.BevelFilter;
import flash.filters.BitmapFilterQuality;
import flash.filters.ColorMatrixFilter;
import flash.filters.ConvolutionFilter;
import flash.filters.ShaderFilter;
import flash.geom.Matrix;
import flash.geom.Matrix3D;
import flash.geom.Point;
import flash.geom.Rectangle;
import flash.geom.Vector3D;
import flash.text.TextField;
import flash.text.TextFieldAutoSize;
import flash.text.TextFormat;
import flash.text.TextLineMetrics;

import DCamera;

[SWF (width="1200", height="900", frameRate="60", backgroundColor="#ffffff")]
public class MipMappedText extends Sprite
{
	[Embed (source = "Assets/Framer.pbj",  mimeType="application/octet-stream")]
	private const FramerPB:Class;

	[Embed (source="C:/Windows/Fonts/pala.ttf", fontName="Palatino_RG_3",
	fontStyle="regular",
	fontWeight="normal",
	mimeType = "application/x-font",
	embedAsCFF=false,
	unicodeRange="U+0021-U+00ff")]
	private const Palatino_RG_3:Class;

	private const IDENTITY:Matrix3D = new Matrix3D();

	private var context3D:Context3D;
	private var program:Program3D;
	private var vertexBuffer:VertexBuffer3D;
	private var indexBuffer:IndexBuffer3D;
	private var texture:Texture;

	private var camera:DCamera;
	private var spm1:SillyPlaneMesh;
	private var spm2:SillyPlaneMesh;

	public function
	MipMappedText()
	{
		stage.align = StageAlign.TOP_LEFT;
		stage.scaleMode = StageScaleMode.NO_SCALE;
		stage.quality = StageQuality.BEST;

		stage.stage3Ds[0].addEventListener (Event.CONTEXT3D_CREATE, initMolehill);
		stage.stage3Ds[0].requestContext3D();
		
		addEventListener (Event.ENTER_FRAME, onRender);
	} // End of Constructor for MipMappedText.
	
	private function
	initMolehill (e:Event)
	:void
	{
		context3D = stage.stage3Ds[0].context3D;			
		context3D.configureBackBuffer (1200, 900, 16, true, true);
		context3D.setCulling ("back");

		var vertices:Vector.<Number> = Vector.<Number> ([
			-1.5,  0.5, 0, 0, 0,
			 1.5,  0.5, 0, 1, 0,
			 1.5, -0.5, 0, 1, 1,
			-1.5, -0.5, 0, 0, 1, 

			-.25,  1.5, 0, 0, 0,
			 .25,  1.5, 0, 1, 0,
			 .25, -1.5, 0, 1, 1,
			-.25, -1.5, 0, 0, 1
			]);
		// Create VertexBuffer3D. 8 vertices, of 5 Numbers each.
		vertexBuffer = context3D.createVertexBuffer (8, 5);
		// Upload VertexBuffer3D to GPU. Offset 0, 8 vertices.
		vertexBuffer.uploadFromVector (vertices, 0, 8);

		var indices:Vector.<uint> = Vector.<uint> ([
			0, 1, 2, 2, 3, 0,
			4, 5, 6, 6, 7, 4
			]);
		// Create IndexBuffer3D. Total of 12 indices. 2 triangles of 3 vertices for 2 planes
		indexBuffer = context3D.createIndexBuffer (12);			
		// Upload IndexBuffer3D to GPU. Offset 0, count 12
		indexBuffer.uploadFromVector (indices, 0, 12);

		/* My apologies for this. Marco's simple example avoids having to distinguish
		** between local and world coordinates, and the example started from that simple premise.
		** Jackson's richer camera requires that level of sophistication. Best of all,
		** Moussa provides some useful mesh classes, but since this is not about building
		** a Molehill framework, I just minimally merged the constructs so we could concentrate
		** on the topic of text as texture without getting too bogged down.
		*/
		spm1 = new SillyPlaneMesh (0, new Vector3D (0.4, -0.6, 0.0), 3.0, 1.0,
			Vector.<Number> ([.8, .8, 0.0, 1]));
		spm2 = new SillyPlaneMesh (6, new Vector3D (0.7, 0.4, -0.01), 0.5, 3.0,
			Vector.<Number> ([0.0, .8, .8, 1]));

		// 2013-06-26 Rectangluar textures are not yet supported, but are due in the next release.
		var maxBBWidth:uint = roundUpMap (400);
		var maxBBHeight:uint = roundUpMap (400);
		texture = context3D.createTexture (maxBBWidth, maxBBHeight, Context3DTextureFormat.BGRA, false);
		// uploadTextureWithMipMaps (maxBBWidth, maxBBHeight, texture, "Moussa\nDembélé");
		uploadTextureWithMipMaps (maxBBWidth, maxBBHeight, texture, "M\nD\nJr");
		// uploadTextureWithMipMaps (maxBBWidth, maxBBHeight, texture, "M");

		// Vertex Constant Register 0 holds the local to world to screen transform.
		// Vertex Constant Register 4 holds scaling and offset data.
		//    If the scaling of the texture to the target is Width-based,
		//    those values are in .x and .y. If the scaling of the texture to the
		//    target is Height-based, those values are in .z and .w.
		// Vertex Attribute Register 0 holds the target object's local position.
		// Vertex Attribute Register 1 holds the UV mapping from target to texture.
		var vertexShaderAssembler:AGALMiniAssembler = new AGALMiniAssembler();
		vertexShaderAssembler.assemble (Context3DProgramType.VERTEX,
			"m44 op, va0, vc0 \n" +
			"mov vt0, va1 \n" +
			"mul vt0.xy, vt0.xy, vc4.xy \n" +
			"add vt0.xy, vt0.xy, vc4.zw \n" +
			"mov v0, vt0"
		);

		// Fragment Constant Register 0 holds 0-1 values to drive true-false logic.
		// Fragment Constant Register 1 holds the Fill Color used to pad the texture.
		var fragmentShaderAssembler:AGALMiniAssembler = new AGALMiniAssembler();
		fragmentShaderAssembler.assemble (Context3DProgramType.FRAGMENT,
			"tex ft1, v0, fs0 <2d, linear, miplinear, clamp> \n" +
			"sge ft2.xy, v0.xy, fc0.xx \n" +
			"mul ft2.x, ft2.x, ft2.y \n" + // if either U or V < 0, ft2.x will be 0.
			"slt ft3.xy, v0.xy, fc0.yy \n" +
			"mul ft3.x, ft3.x, ft3.y \n" + // if either U or V > 1, ft3.x will be 0.
			"mul ft4.x, ft2.x, ft3.x \n" + // So, if U or V out of bounds, ft4.x will be 0.
			"mul ft1.xyz, ft1.xyz, ft4.xxx \n" + // If UV in bounds, original RGB preserved, else 0.
			"sub ft5.x, fc0.y, ft4.x \n" + // If UV was out of bounds,
			"mul ft6.xyz, fc1.xyz, ft5.xxx \n" + // ft6 gets background color, else zero.
			"add ft6.xyz, ft6.xyz, ft1.xyz \n" + // This assumes out of bounds left interpolated color zero.
			"mov ft6.w, fc1.w \n" +
			"mov oc, ft6"
		);
		program = context3D.createProgram();
		program.upload (vertexShaderAssembler.agalcode, fragmentShaderAssembler.agalcode);

		camera = new DCamera (stage, 0.1, 100.0,
			(stage.stageWidth / stage.stageHeight), (40 * Math.PI / 180.0),
			0, 0, 2, // position
			0, 0, 0, // target
			0, 1, 0 // up dir
			);
	} // End of initMolehill().

	public function
	uploadTextureWithMipMaps (maxW:uint, maxH:uint, texture:Texture, text:String)
	:void 
	{
		var bmd:BitmapData;
		var mapW:uint = maxW;
		var mapH:uint = maxH;
		var mipLevel:uint = 0;
		while ((mapW > 0) && (mapH > 0)) {
			bmd = paintText (mapW, mapH, text, 0xfffff000,  mipLevel);
			texture.uploadFromBitmapData (bmd, mipLevel);

			// Texture map added to stage just for debugging purposes.
			var bm:Bitmap = new Bitmap (bmd);
			bm.x = (10 + (mipLevel * 100.0));
			bm.y = (10 + (mipLevel * 100.0));
			stage.addChild (bm);

			mapW >>>= 1;
			mapH >>>= 1;
			++ mipLevel;
		}
	} // End of uploadTextureWithMipMaps().

	private function
	paintText (mapW:uint, mapH:uint, text:String, fillColor:uint, mipLevel:uint)
	:BitmapData
	{
		var bmd:BitmapData = new BitmapData (mapW, mapH, true, (fillColor |  (0xff113377 << mipLevel)));
		var tf:TextField = new TextField();
		tf.embedFonts = true;
		tf.multiline = true; 
		tf.autoSize = TextFieldAutoSize.LEFT;

		var layoutHorizontal:Boolean = true;
		var textSize:int = autoSizeFont (mapW, mapH, text);
		if (textSize < 0) {
			layoutHorizontal = false;
			textSize = -textSize;
		}
		// trace ("Mip Level", mipLevel, "Map Width", mapW, "Map Height", mapH, "Text Size", textSize);

		var rgTF:TextFormat = new TextFormat ("Palatino_RG_3", textSize, 0x000000);
		tf.defaultTextFormat = rgTF;
		tf.appendText (text);
		var metrics:TextLineMetrics = tf.getLineMetrics (0);
		tf.replaceText (0, tf.length, (text + "\n" + mipLevel + " :: " + textSize));
		if (layoutHorizontal) {
			rgTF.leading = -(metrics.descent * 0.9);
		} else {
			rgTF.leading = -(metrics.descent * 1.2);
		}
		tf.setTextFormat (rgTF);

		tf.blendMode = BlendMode.MULTIPLY;
		var bf:BevelFilter = new BevelFilter (2, 15, 0x111111, 1, 0xaaaaaa, 1,
			2, 2, 200, BitmapFilterQuality.HIGH, "inner", false);
		var cf:ConvolutionFilter = new ConvolutionFilter (3, 3, [ -1, 0, -1, 0, 4, 0, -1, 0, -1]);
		tf.filters = [cf, bf];

		var mat:Matrix = new Matrix();
		if (layoutHorizontal) {
			mat.translate ((textSize * 0.1), -(metrics.descent / 2));
		} else {
			if (text.length == 1) {
				mat.translate ((textSize * 0.02), -(metrics.descent * 0.7));
			} else {
				mat.translate ((textSize * 0.08), -metrics.descent);
			}
		}
		bmd.drawWithQuality (tf, mat, null, null, null, true, StageQuality.BEST);

		var framer:Shader = new Shader();
		framer.byteCode = new FramerPB();
		framer.data.srcSize.value = [bmd.width, bmd.height];
		framer.data.width.value = [3];
		framer.data.frameColor.value = [0.0, 0.0, 0.0, 1.0];
		bmd.applyFilter (bmd, bmd.rect, new Point (0, 0), new ShaderFilter (framer));

		return (bmd);
	} // End of paintText().

	private function
	onRender (e:Event)
	:void
	{
		if (! context3D) return;
		
		context3D.clear (.8, .8, .8);

		var world2Clip:Matrix3D = camera.worldToClipMatrix;
		var concatMat:Matrix3D = IDENTITY;
	
		// Vertex position to attribute register 0
		context3D.setVertexBufferAt (0, vertexBuffer, 0, Context3DVertexBufferFormat.FLOAT_3);
		// UV to attribute register 1
		context3D.setVertexBufferAt (1, vertexBuffer, 3, Context3DVertexBufferFormat.FLOAT_2);
		// Assign shader program
		context3D.setProgram (program);				

		context3D.setTextureAt (0, texture);

		var iar:Number; // Inverse Aspect Ratio.
		var off:Number = ((iar - 1) / 2); // Offset.
		var vc4:Vector.<Number>;
		if ((spm1.width / spm1.height) > 1.0) {
			iar = (spm1.width / spm1.height);
			off =  ((iar - 1) / 2);
			vc4 = Vector.<Number> ([iar, 1, -off, 0]);
		} else {
			iar = (spm1.height / spm1.width);
			off =  ((iar - 1) / 2);
			vc4 = Vector.<Number> ([1, iar, 0, -off]);
		}
		context3D.setProgramConstantsFromVector (Context3DProgramType.VERTEX, 4, vc4);
		context3D.setProgramConstantsFromVector (Context3DProgramType.FRAGMENT, 0, Vector.<Number> ([0, 1, 0, 0]));
		context3D.setProgramConstantsFromVector (Context3DProgramType.FRAGMENT, 1, spm1.fillColor);

		spm1.mat.copyToMatrix3D (concatMat);
		concatMat.prepend (world2Clip);
		context3D.setProgramConstantsFromMatrix (Context3DProgramType.VERTEX, 0, concatMat, false);
		context3D.drawTriangles (indexBuffer, spm1.ibOffset, 2);

		if ((spm2.width / spm2.height) > 1.0) {
			iar = (spm2.width / spm2.height);
			off =  ((iar - 1) / 2);
			vc4 = Vector.<Number> ([iar, 1, -off, 0]);
		} else {
			iar = (spm2.height / spm2.width);
			off =  ((iar - 1) / 2);
			vc4 = Vector.<Number> ([1, iar, 0, -off]);
		}
		context3D.setProgramConstantsFromVector (Context3DProgramType.VERTEX, 4, vc4);
		context3D.setProgramConstantsFromVector (Context3DProgramType.FRAGMENT, 0, Vector.<Number> ([0, 1, 0, 0]));
		context3D.setProgramConstantsFromVector (Context3DProgramType.FRAGMENT, 1, spm2.fillColor);

		spm2.mat.copyToMatrix3D (concatMat);
		concatMat.prepend (world2Clip);
		context3D.setProgramConstantsFromMatrix (Context3DProgramType.VERTEX, 0, concatMat, false);
		context3D.drawTriangles (indexBuffer, spm2.ibOffset, 2);

		context3D.setVertexBufferAt (0, null);
		context3D.setVertexBufferAt (1, null);
		context3D.setTextureAt (0, null);

		context3D.present();
	} // End of onRender().

	private function
	roundUpMap (value:uint)
	:uint
	{
		var pow2:uint = 1;
		while (pow2 < value) pow2 <<= 1;
		return (pow2);
	} // End of roundUpMap().

	private function
	autoSizeFont (mapW:uint, mapH:uint, text:String)
	:int
	{
		var lines:Array = text.split ("\n");
		var numLines:uint = lines.length;
		var widest:uint = 0;
		for each (var line:String in lines) {
			if (line.length > widest) widest = line.length;
		}
		var result:int;
		if (widest > numLines) {
			result = Math.floor (mapW / widest);
		} else {
			result = -Math.floor (mapH / numLines);
		}
		return (result);
	} // End of autoSizeFont().

} // End of MipMappedText Class.

} // End of Package Declaration.

import flash.geom.Matrix3D;
import flash.geom.Vector3D;

internal final class SillyPlaneMesh
{
	public var ibOffset:uint;
	public var mat:flash.geom.Matrix3D;
	public var width:Number;
	public var height:Number;
	public var fillColor:Vector.<Number>;

	public function
	SillyPlaneMesh (ibOffset:uint, worldCenter:Vector3D, width:Number, height:Number,
		fillColor:Vector.<Number>)
	{
		this.ibOffset = ibOffset;
		mat = new Matrix3D (new <Number>[
			1.0, 0.0, 0.0, worldCenter.x,
			0.0, 1.0, 0.0, worldCenter.y,
			0.0, 0.0, 1.0, worldCenter.z,
			0.0, 0.0, 0.0, 1.0
			]);
		this.width = width;
		this.height = height;
		this.fillColor = fillColor;
	} // End of Constructor for SillyPlaneMesh.

} // End of Internal SillyPlaneMesh Class.
